/*
 * Copyright (c) 2010 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-dkim-0.c,v 1.33 2013/12/03 02:52:52 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/stat.h"
#include "sm/fcntl.h"
#include "sm/io.h"
#include "sm/string.h"
#include "sm/mta.h"
#include "sm/misc.h"
#include "sm/base64.h"
#include "sm/dkimdef.h"
#include "sm/dkim.h"
#include "sm/sysexits.h"
#include "sm/test.h"

#include "dkim.h"

#if MTA_USE_DKIM
#define SIG "v=1; a=rsa-sha256; c=simple/simple; d=example.net;\r\n\ts=default; t=1258487282;\r\n\tbh=P7M5AsnZKX0CSpQl3w3u9XwzYQ7gDrSE5gSSQy/1W2E=; h=From:To:Subject;\r\n\tb=bJ2PgR07u9NdHD8UgExfjc6UGg0PD8Ac7yg1YlSlWFSW1oboC4jUnzBTsCFE6Gqml\r\n\t yQISCypKClJt1746n3UYa61bOkcTiEeD27AIX+bbEXWqyw/+Y9RNHehcYT7eY1ITG3\r\n\t C1xeoJ2aNeiEPcG2SPYf8JsuY8dzzeOGe7CONvGs="

#define BH "P7M5AsnZKX0CSpQl3w3u9XwzYQ7gDrSE5gSSQy/1W2E="

const char *bodyhash64 = BH;

/*
BHH SHA1: 4a7bcaf2d773dfafeb5fd31ef1d80a4819d99f60
*/

#define BODY "2020 Vision Street\r\nAnycity, Anystate 90009\r\nOffice ph: (510) 555-1234\r\nCell ph:   (510) 555-5678\r\nfax:       (510) 555-9000\r\nemail: mmtest3@v253.pogonip.net\r\n"

static uint Verbose = 0;
static bool Bin = false;
static dkim_ctx_T dkim_ctx;
static dkim_msg_P dkim_msg;
static dkim_s_T dkim_s;

static void
printhash(const char *s, const uchar *hash, size_t hlen, uchar *b64hash, size_t bhlen)
{
	size_t i;

	if (Verbose > 0 && s != NULL && *s != '\0')
		sm_io_fputs(smioout, (const uchar *)s);
	if (Verbose > 1) {
		for (i = 0; i < hlen; i++)
			if (Bin)
				sm_io_fprintf(smioout, "%c", hash[i]);
			else
				sm_io_fprintf(smioout, "%02x", hash[i]);
		if (!Bin) sm_io_fprintf(smioout, "\n");
	}
	sm_memzero(b64hash, bhlen);
	sm_base64_encode(hash, hlen, b64hash, bhlen);
	if (Verbose > 0) {
		sm_io_fprintf(smioout, "%s\n", b64hash);
		sm_io_flush(smioout);
	}
}

/*
**  T_DKIM_MSG -- compute msg hashes
**
**	Parameters:
**		msg -- msg
**		len -- len of msg
**		gensig -- generate a signature?
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
t_dkim_msg(const uchar *msg, size_t len, bool gensig)
{
	sm_ret_T ret;
	size_t i, hashlen;
	ssize_t bc;
	uchar *hash;
	uchar bhash[SHA256_DIGEST_LENGTH];
	uchar b64hash[256];
	dkim_cookie_P dkim_cookie;
	uchar c;
	uint eot_state, eoh_state;
	sm_file_T *hfp, *bfp;
	static SM_DECL_EOT;
	static SM_DECL_EOH;
#define SM_CRLFCR	(SM_EOH_LEN - 1)

	ret = dkim_open_write(dkim_msg, DKIM_SM_RSA_SHA256, &hfp);
	SM_TEST_E(SM_SUCCESS == ret);
	dkim_cookie = (dkim_cookie_P)f_cookie(*hfp);

	/* fixed time (magic value) */
	dkim_msg->dkm_time = 1258487282;

	eoh_state = 0;
	eot_state = SM_GOT_CRLF;
	bc = 0;
	for (i = 0; i < len; i++) {
		c = msg[i];

		/* headers? */
		if (eoh_state < SM_EOH_LEN) {
			if (c == eoh[eoh_state]) {
				++eoh_state;
				if (SM_EOH_LEN == eoh_state) {

					/*
					 *  did I put too much into the buffer?
					 *  \r\n from empty line?
					 */


					/* open DKIM buffer for body */
					ret = dkim_open_write(dkim_msg, DKIM_SM_RSA_SHA256, &bfp);
					SM_TEST_E(SM_SUCCESS == ret);
					continue;
				}
				if (eoh_state == SM_CRLFCR)
					continue;
			}
			else {
				if (eoh_state == SM_CRLFCR)
					ret = sm_io_putc(hfp, '\r');
				eoh_state = 0;
				if (c == eoh[eoh_state])
					++eoh_state;
			}
			ret = sm_io_putc(hfp, c);
			continue;
		}

		/* need to ignore trailing empty lines */
		if (c == eot[eot_state]) {
			if (++eot_state>= SM_EOT_LEN)
				break;
		}
		else {
			eot_state = 0;
			if (c == eot[eot_state])
				++eot_state;
		}

		/* skip leading '.' */
		if (eot_state == 3 && c == '.')
			continue;
		ret = sm_io_putc(bfp, c);
		++bc;
	}

	if (Verbose > 2) {
		sm_io_fprintf(smioout, "len=%lu, bc=%lu\n", len, bc);
		sm_io_flush(smioout);
	}
	dkim_gethash(bfp, &hash, &hashlen);
	sm_memcpy(bhash, hash, sizeof(bhash));
	printhash( "body-hash:", bhash, SHA256_DIGEST_LENGTH, b64hash, sizeof(b64hash));
	if (bodyhash64 != NULL && *bodyhash64 != '\0')
		SM_TEST_E(sm_streq(bodyhash64, (const char *)b64hash));
	ret = dkim_close(bfp, 0);
	SM_TEST(SM_SUCCESS == ret);

	if (gensig && dkim_s.dks_rsa != NULL) {

		/*
		 *  need to create a DKIM header with all data in it
		 *  except for b= of course.
		 *  (that needs to be a library function)
		 */

		sm_io_fprintf(hfp,
			"DKIM-Signature: v=1; a=rsa-sha256; c=simple/simple; d=example.net;\r\n\ts=default; t=%v;\r\n\tbh=%s; h=From:To:Subject;\r\n\tb=",
			dkim_msg->dkm_time, b64hash);

		dkim_gethash(hfp, &hash, &hashlen);
		sm_memcpy(bhash, hash, sizeof(bhash));
		printhash("\nheader-hash:", bhash, SHA256_DIGEST_LENGTH, b64hash, sizeof(b64hash));

/*
v=1; a=rsa-sha256; c=simple/simple; d=example.net;
	s=default; t=1258487282;
	bh=P7M5AsnZKX0CSpQl3w3u9XwzYQ7gDrSE5gSSQy/1W2E=; h=From:To:Subject;
	b=LMogMTvrbyx4VeI4Yqw5K17rdFBTuMmuFGWKfJDxFHuVirOoOAA9SYDFucEYQVhOd
	 gwqlVbJszGxQw0PMNqvwdWYRmQ5JKV7Uq+3oRImRX+zmGr5ZiHQxYXYSV5T5m0SQx3
	 HQjB3qM3DiVHVAovhBgqDFfHqglC0jxVfITMQP2w=
*/

		dkim_msg->dkm_s = &dkim_s;
		ret = dkim_gensig(&dkim_ctx, dkim_msg, dkim_cookie);
		SM_TEST_E(SM_SUCCESS == ret);

		sm_memzero(b64hash, sizeof(b64hash));
		sm_base64_encode(dkim_msg->dkm_rsaout,
			dkim_msg->dkm_rsaoutlen, b64hash, sizeof(b64hash));
		if (Verbose > 0) {
			sm_io_fprintf(smioout, "sig: %s\n", b64hash);
			sm_io_flush(smioout);
		}

		ret = dkim_close(hfp, 0);
		SM_TEST(SM_SUCCESS == ret);
	}
	else {
		ret = dkim_close(hfp, 0);
		SM_TEST(SM_SUCCESS == ret);
	}

	return SM_SUCCESS;

  error:
	return ret;
}

/*
**  DKIM_BODY -- compute hash of body
**
**	Parameters:
**		dfp -- DKIM FILE
**		parseit -- parse the body (or treat it as a blob)?
**		body -- body
**		len -- len of body
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
dkim_body(sm_file_T *dfp, bool parseit, const uchar *body, size_t len)
{
	sm_ret_T ret;
	size_t i, hashlen;
	ssize_t byteswritten;
	uchar *hash;
	uchar b64hash[256];
	dkim_cookie_P dkim_cookie;
	SM_DECL_EOT;

	dkim_cookie = (dkim_cookie_P)f_cookie(*dfp);

	/* write some stuff ... */

	if (parseit) {
		uchar c;
		uint eot_state, eoh_state;

		eoh_state = 0;
		eot_state = SM_GOT_CRLF;
		for (i = 0; i < len; i++) {
			c = body[i];
			if (c == eot[eot_state]) {
				if (++eot_state>= SM_EOT_LEN)
					break;
			}
			else {
				eot_state = 0;
				if (c == eot[eot_state])
					++eot_state;
			}

			/* need to ignore trailing empty lines */

			/* skip leading '.' */
			if (eot_state == 3 && c == '.')
				continue;
			ret = sm_io_putc(dfp, c);
		}
	}
	else {
		ret = dkim_write(dfp, (uchar *) body, len, &byteswritten);
		SM_TEST_E(SM_SUCCESS == ret);
		SM_TEST(byteswritten == len);
	}

	dkim_gethash(dfp, &hash, &hashlen);

	printhash(NULL, hash, SHA256_DIGEST_LENGTH, b64hash, sizeof(b64hash));
	SM_TEST(sm_streq(bodyhash64, (const char *)b64hash));
	if (!sm_streq(bodyhash64, (const char *)b64hash) && Verbose > 0)
		sm_io_fprintf(smioerr, "error: bodyhash64=%s, b64hash=%s\n",
			bodyhash64, b64hash);
	return SM_SUCCESS;

  error:
	return ret;
}

/*
**  DKIM_TEST -- simple DKIM test
**
**	Parameters:
**		body -- body
**		len -- len of body
**		parseit -- parse the body?
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
dkim_test(const uchar *body, size_t len, bool parseit)
{
	sm_ret_T ret;
	sm_file_T *dfp;

	ret = dkim_open_write(dkim_msg, DKIM_SM_RSA_SHA256, &dfp);
	SM_TEST_E(SM_SUCCESS == ret);

	/* write some stuff ... */
	ret = dkim_body(dfp, parseit, body, len);
	SM_TEST_E(SM_SUCCESS == ret);

	ret = dkim_close(dfp, 0);
	SM_TEST(SM_SUCCESS == ret);

  error:
	return ret;
}

/*
**  USAGE -- usage message
**
**	Parameters:
**		prg -- program name
**
**	Returns:
**		none
*/

static void
usage(char *prg)
{
	sm_io_fprintf(smioerr, "usage: %s [options]\n", prg);
	return;
}
#endif /* MTA_USE_DKIM */

int
main(int argc, char *argv[])
{
#if MTA_USE_DKIM
	int c, r, loops;
	size_t msglen;
	bool parseit, gensig;
	uchar *body, *msg;
	dkim_cnf_T dkim_cnf;

	body = (uchar *)BODY;
	msglen = strlen((const char *)body);
	msg = NULL;
	loops = 0;
	parseit = gensig= false;
	sm_memzero(&dkim_ctx, sizeof(dkim_ctx));
	sm_memzero(&dkim_s, sizeof(dkim_s));
	sm_memzero(&dkim_cnf, sizeof(dkim_cnf));
	dkim_ctx.dkc_cnf = &dkim_cnf;
	while ((c = getopt(argc, argv, "b:Cf:h:k:l:m:psV")) != -1) {
		switch (c) {
		  case 'b':
			body = (uchar *)optarg;
			msglen = strlen((const char *)body);
			break;
		  case 'C':
			compoptions(smioout);
			return 0;
		  case 'f':
			r = readfile(optarg, &body, &msglen);
			if (r != 0)
				return r;
			break;
		  case 'h':
			bodyhash64 = optarg;
			break;
		  case 'k':
			dkim_s.dks_keyfile = optarg;
			r = dkim_readkey(&dkim_s);
			if (r != 0)
				return r;
			break;
		  case 'l':
			loops = strtol(optarg, NULL, 0);
			break;
		  case 'm':
			r = readfile(optarg, &msg, &msglen);
			if (r != 0)
				return r;
			break;
		  case 'p':
			parseit = true;
			break;
		  case 's':
			gensig = true;
			break;
		  case 'V':
			++Verbose;
			break;
		  default:
			usage(argv[0]);
			return EX_USAGE;
		}
	}

	sm_test_begin(argc, argv, "test dkim 0");
	r = dkim_msg_new(&dkim_ctx, &dkim_msg);
	SM_TEST_E(SM_SUCCESS == r);
	if (msg != NULL) {
		if (loops > 0) {
			time_t start;

			start = time((time_t *) 0);
			for (c = 0; c < loops; c++) {
				r = t_dkim_msg(msg, msglen, gensig);
				if (r != SM_SUCCESS)
					break;
			}
			sm_io_fprintf(smioerr, "elapsed time: %ld\n",
				(long) time((time_t *) 0) - start);
		}
		else
			t_dkim_msg(msg, msglen, gensig);
	}
	else
		dkim_test(body, msglen, parseit);
  error:
	return sm_test_end();
#else
	sm_io_fprintf(smioerr, "%s: SKIPPED: MTA_USE_DKIM not set\n", argv[0]);
	return 0;
#endif
}
