/*
 * Copyright (c) 2002-2006 Sendmail, Inc. and its suppliers.
 *      All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: dnstsk.c,v 1.120 2013/12/27 05:03:20 ca Exp $")
#include "sm/ctype.h"
#include "sm/io.h"	/* sm_fd_nonblock() */
#include "sm/dns.h"
#include "sm/dns-int.h"
#include "sm/dnstsk.h"
#include "dns.h"
#define LIBDNS_LOG_DEFINES 1
#include "log.h"

/*
**  Note: for some unknown reasons the query in the answer packet
**  does not contain a trailing dot even if the original query did.
**  Hence the code that generates a key removes trailing dots.
**  ToDo: It would be nice to put that key generation code into one place
**  (macro/function).
*/

/* when to wakeup dns_tsk_cleanup; should be "never" by default, see comments */
#define TSK_CLEANUP_SLP	60
#define TSK_CLEANUP_SLP_MIN	1
static sm_ret_T	 dns_tsk_cleanup(sm_evthr_task_P _tsk);

#if !MTA_USE_PTHREADS

/*
**  ToDo: Build a version for state threads... add macros to deal with the
**  differences between event threads and state threads...
*/

/* HACK */
# define EVTHR_DEL	1
# define EVTHR_WAITQ	0
# define sm_evthr_task_P void *
#endif /* !MTA_USE_PTHREADS */

#ifndef NS_OPT_DNSSEC_OK
# define NS_OPT_DNSSEC_OK   0x8000U
#endif

#if MTA_USE_DNSSEC
/*
**  NS_NMKQUERY -- create a DNS query
**
**	Parameters:
**		dname --  domain name
**		class -- class of query
**		type --  type of query
**		*buf --  buffer to put query
**		buflen -- size of buffer
**
**	Return value:
**		-1: error
**		>=0: number of elements in buf
*/

static int
ns_nmkquery(const char *dname, int class, int type, uchar *buf, int buflen)
{
	HEADER *hp;
	uchar *cp, *ep;
	int n;
	uchar *dnptrs[20], **dpp, **lastdnptr;

	if ((buf == NULL) || (buflen < HFIXEDSZ))
		return (-1);
	memset(buf, 0, HFIXEDSZ);
	hp = (HEADER *) buf;
	hp->id = 1579;
	hp->opcode = QUERY;
	hp->rd = 1;
	hp->rcode = NOERROR;
	cp = buf + HFIXEDSZ;
	ep = buf + buflen;
	dpp = dnptrs;
	*dpp++ = buf;
	*dpp++ = NULL;
	lastdnptr = dnptrs + sizeof dnptrs / sizeof dnptrs[0];
	if (ep - cp < QFIXEDSZ)
		return (-1);
	if ((n = dn_comp(dname, cp, ep - cp - QFIXEDSZ, dnptrs, lastdnptr)) < 0)
		return (-1);
	cp += n;
	PUTSHORT(type, cp);
	PUTSHORT(class, cp);
	hp->qdcount = htons(1);

	return (cp - buf);
}

# ifdef RES_USE_EDNS0
/* attach OPT pseudo-RR, as documented in RFC2671 (EDNS0). */
#  ifndef T_OPT
#   define T_OPT	41
#  endif

/*
**  NS_NOPT -- add more DNS options
**
**	Parameters:
**		n0 -- current offset in buffer
**		buf -- buffer to put query
**		buflen -- size of buffer
**		anslen -- UDP answer buffer size
**
**	Return value:
**		-1: error
**		>=0: number of elements in buf
*/

static int
ns_nopt(int n0, uchar *buf, int buflen, int anslen)
{
	HEADER *hp;
	uchar *cp, *ep;
	uint16_t flags;

	hp = (HEADER *) buf;
	cp = buf + n0;
	ep = buf + buflen;

	if ((ep - cp) < 1 + RRFIXEDSZ)
		return (-1);

	*cp++ = 0;	/* "." */

	PUTSHORT(T_OPT, cp);	/* TYPE */
	PUTSHORT(anslen & 0xffff, cp);	/* CLASS = UDP payload size */
	*cp++ = NOERROR;	/* extended RCODE */
	*cp++ = 0;		/* EDNS version */
	flags = NS_OPT_DNSSEC_OK;
	PUTSHORT(flags, cp);
	PUTSHORT(0, cp);	/* RDLEN */
	hp->arcount = htons(ntohs(hp->arcount) + 1);

	return (cp - buf);
}
# endif /* RES_USE_EDNS0 */
#endif /* MTA_USE_DNSSEC */

/*
**  ToDo:
**	- fix hash key (query + type); done, but it's still ugly.
**	- deal with partial reads (requeue, store state in dns_tsk)
*/

/*
**  DNS_TSK_DEL -- delete a DNS task
**
**	Parameters:
**		dns_tsk -- DNS task context
**
**	Return value:
**		usual sm_error code
*/

sm_ret_T
dns_tsk_del(dns_tsk_P dns_tsk)
{
	if (NULL == dns_tsk)
		return SM_SUCCESS;
	SM_IS_DNSTSK(dns_tsk);
	if (is_valid_socket(dns_tsk->dnstsk_fd)) {
		close(dns_tsk->dnstsk_fd);
		dns_tsk->dnstsk_fd = INVALID_SOCKET;
	}
	SM_STR_FREE(dns_tsk->dnstsk_wr);
	SM_STR_FREE(dns_tsk->dnstsk_rd);
#if MTA_USE_PTHREADS
	if (DNS_TSK_IS_FLAG(dns_tsk, DNS_TSK_FL_HAS_MTX)) {
		DNS_TSK_CLR_FLAG(dns_tsk, DNS_TSK_FL_HAS_MTX);
		(void) pthread_mutex_destroy(&dns_tsk->dnstsk_mutex);
	}
#endif
	dns_tsk->sm_magic = SM_MAGIC_NULL;
	sm_free_size(dns_tsk, sizeof(*dns_tsk));
	return SM_SUCCESS;
}

/*
**  DNS_CONNECT -- 
**
**	Parameters:
**		dns_tsk -- DNS task context
**
**	Return value:
**		usual sm_error code
*/

static sm_ret_T
dns_connect(dns_tsk_P dns_tsk, uint flags)
{
	sm_ret_T ret;

	dns_tsk->dnstsk_fd = socket(dns_tsk->dnstsk_sin.sin_family,
				(flags & DNS_TSK_FL_USETCP) ? SOCK_STREAM : SOCK_DGRAM, 0);
	if (!is_valid_socket(dns_tsk->dnstsk_fd)) {
		ret = sm_error_temp(SM_EM_DNS, errno);
		goto error;
	}
	if ((flags & (DNS_TSK_FL_USETCP|DNS_TSK_FL_CONNECTUDP)) != 0) {
		if (connect(dns_tsk->dnstsk_fd, (sockaddr_P) &dns_tsk->dnstsk_sin,
			sizeof(sockaddr_in_T)) != 0)
		{
			ret = sm_error_temp(SM_EM_DNS, errno);
			goto error;
		}
	}
	ret = sm_fd_nonblock(dns_tsk->dnstsk_fd, true);
	if (sm_is_err(ret))
		goto error;
	return ret;

  error:
	return ret;
}

/*
**  DNS_RECONNECT --  reconnect to a DNS server
**
**	Parameters:
**		tsk -- evthr task
**
**	Return value:
**		usual sm_error code
*/

static sm_ret_T
dns_reconnect(sm_evthr_task_P tsk)
{
	sm_ret_T ret;
	dns_tsk_P dns_tsk;

	SM_IS_EVTHR_TSK(tsk);
	dns_tsk = (dns_tsk_P) tsk->evthr_t_actx;
	SM_IS_DNSTSK(dns_tsk);
	if (is_valid_fd(tsk->evthr_t_fd)) {
		close(tsk->evthr_t_fd);
		tsk->evthr_t_fd = dns_tsk->dnstsk_fd = INVALID_SOCKET;
	}
	ret = dns_connect(dns_tsk, dns_tsk->dnstsk_flags);
	MTA_LIBDNS_DBG_DPRINTF((smioerr,
		"sev=DBG, func=dns_reconnect, tsk=%p, dns_tsk=%p, ret=%r, fd=%d\n",
		tsk, dns_tsk, ret, dns_tsk->dnstsk_fd));
	if (sm_is_err(ret))
		goto error;
	DNS_TSK_CLR_FLAG(dns_tsk, DNS_TSK_FL_EOF);
	tsk->evthr_t_fd = dns_tsk->dnstsk_fd;
	return ret;

  error:
	return ret;
}

/*
**  DNS_TSK_NEW -- create a new dns_tsk
**
**	Parameters:
**		dns_mgr_ctx -- DNS manager context
**		flags -- flags
**		ipv4 -- IPv4 address of DNS resolver
**		port -- port of DNS resolver (0: use default)
**		pdns_tsk -- (pointer to) DNS task context (output)
**
**	Return value:
**		usual sm_error code
*/

sm_ret_T
dns_tsk_new(dns_mgr_ctx_P dns_mgr_ctx, uint flags, uint32_t ipv4, ushort port, dns_tsk_P *pdns_tsk)
{
	sm_ret_T ret;
	uint u;
	dns_tsk_P dns_tsk;
#if MTA_USE_PTHREADS
	int r;
#endif

	SM_IS_DNSM_CTX(dns_mgr_ctx);
	SM_REQUIRE(pdns_tsk != NULL);

	u = dns_mgr_ctx->dnsmgr_ntsks;
	if (u >= SM_DNS_TSKS_MAX)
		return sm_error_perm(SM_EM_DNS, SM_E_FULL);
	dns_tsk = (dns_tsk_P) sm_zalloc(sizeof(*dns_tsk));
	if (NULL == dns_tsk)
		return sm_error_temp(SM_EM_DNS, ENOMEM);
	dns_tsk->sm_magic = SM_DNST_MAGIC;
	dns_tsk->dnstsk_fd = INVALID_SOCKET;
	dns_tsk->dnstsk_flags = flags & DNS_TSK_FL_OPT_MSK;
	dns_tsk->dnstsk_mgr = dns_mgr_ctx;
	dns_tsk->dnstsk_wr = sm_str_new(NULL, HFIXEDSZ + MAXPACKET, QUERY_SIZE_MAX);
	if (NULL == dns_tsk->dnstsk_wr) {
		ret = sm_error_temp(SM_EM_DNS, ENOMEM);
		goto error;
	}
	dns_tsk->dnstsk_rd = sm_str_new(NULL, HFIXEDSZ + MAXPACKET, QUERY_SIZE_MAX);
	if (NULL == dns_tsk->dnstsk_rd) {
		ret = sm_error_temp(SM_EM_DNS, ENOMEM);
		goto error;
	}

	/* Set up the socket description */
	sm_memzero(&dns_tsk->dnstsk_sin, sizeof(dns_tsk->dnstsk_sin));
	dns_tsk->dnstsk_sin.sin_family = AF_INET;
	dns_tsk->dnstsk_sin.sin_port = htons(port != 0 ? port : DNS_PORT);
	sm_memcpy(&dns_tsk->dnstsk_sin.sin_addr.s_addr, &ipv4, sizeof(ipv4));

#if MTA_USE_PTHREADS
	r = pthread_mutex_init(&dns_tsk->dnstsk_mutex, SM_PTHREAD_MUTEXATTR);
	if (r != 0) {
		ret = sm_error_perm(SM_EM_DNS, r);
		goto error;
	}
	DNS_TSK_SET_FLAG(dns_tsk, DNS_TSK_FL_HAS_MTX);
#endif /* MTA_USE_PTHREADS */

	ret = dns_connect(dns_tsk, flags);
	if (sm_is_err(ret))
		goto error;
	DNSTRQL_INIT(dns_tsk);
	dns_tsk->dnstsk_idx = u;
	dns_tsk->dnstsk_maxtimeouts = DNS_MAXTIMEOUTS;
	dns_mgr_ctx->dnsmgr_dnstsks[u] = dns_tsk;
	dns_mgr_ctx->dnsmgr_tskstatus[u] = DNSTSK_ST_INIT;
	++dns_mgr_ctx->dnsmgr_ntsks;

	*pdns_tsk = dns_tsk;
	return SM_SUCCESS;

  error:
	(void) dns_tsk_del(dns_tsk);
	return ret;
}

/*
**  DNS_TSK_START -- start all(?) DNS tasks
**
**	Parameters:
**		dns_mgr_ctx -- DNS manager context
**		evthr_ctx -- Event thread system context
**
**	Return value:
**		usual sm_error code
*/

sm_ret_T
dns_tsk_start(dns_mgr_ctx_P dns_mgr_ctx, sm_evthr_ctx_P evthr_ctx)
{
	sm_ret_T ret;
	uint u;
	sm_evthr_task_P	task;
	timeval_T sleept;

	SM_IS_DNSM_CTX(dns_mgr_ctx);
	for (u = 0; u < dns_mgr_ctx->dnsmgr_ntsks; u++) {
		task = NULL;
		ret = evthr_task_new(evthr_ctx, &task, EVTHR_EV_RD,
				dns_mgr_ctx->dnsmgr_dnstsks[u]->dnstsk_fd, NULL, dns_comm_tsk,
				(void *) dns_mgr_ctx->dnsmgr_dnstsks[u]);
		if (sm_is_err(ret))
			goto error;
		dns_mgr_ctx->dnsmgr_tsk[u] = task;
		dns_mgr_ctx->dnsmgr_tskstatus[u] = DNSTSK_ST_OK;
	}

	ret = evthr_timeval(evthr_ctx, &sleept);
	sleept.tv_sec += TSK_CLEANUP_SLP;
	ret = evthr_task_new(evthr_ctx, &task, EVTHR_EV_SL, INVALID_FD, &sleept,
				dns_tsk_cleanup, (void *) dns_mgr_ctx);
	if (sm_is_err(ret))
		goto error;
	dns_mgr_ctx->dnsmgr_cleanup = task;
	return ret;

  error:
	/* Is this ok?? How to properly terminate task?? */
	for (u = 0; u < dns_mgr_ctx->dnsmgr_ntsks; u++) {
		task = dns_mgr_ctx->dnsmgr_tsk[u];
		if (task != NULL) {
			evthr_task_del(evthr_ctx, task, THR_LOCK_UNLOCK);
			dns_mgr_ctx->dnsmgr_tsk[u] = NULL;
		}
		dns_mgr_ctx->dnsmgr_tskstatus[u] = DNSTSK_ST_NONE;
	}
	task = dns_mgr_ctx->dnsmgr_cleanup;
	if (task != NULL) {
		evthr_task_del(evthr_ctx, task, THR_LOCK_UNLOCK);
		dns_mgr_ctx->dnsmgr_cleanup = NULL;
	}
	return ret;
}

/*
**  DNS_REQ_FREE -- free a DNS request,
**		remove it from list if dns_mgr_ctx is not NULL.
**
**	Parameters:
**		dns_req -- DNS request
**		dns_mgr_ctx -- DNS manager context
**
**	Return value:
**		usual sm_error code
**
**	Locking: dns_mgr_ctx must be locked by caller.
*/

static sm_ret_T
dns_req_free(dns_req_P dns_req, dns_mgr_ctx_P dns_mgr_ctx)
{
	MTA_LIBDNS_DBG_DPRINTF((smioerr,
		"sev=DBG, func=dns_req_free, dns_req=%p, ctx=%p, inlist=%d\n",
		dns_req, dns_mgr_ctx,
		NULL == dns_mgr_ctx ? -1 :  DNSREQ_IS_INANYLIST(dns_req)));
	if (NULL == dns_req)
		return SM_SUCCESS;
	MTA_LIBDNS_DBG_DPRINTF((smioerr,
		"sev=DBG, func=dns_req_free, dns_req=%p, key=%p, query=%p\n",
		dns_req, dns_req->dnsreq_key, dns_req->dnsreq_query));
	SM_CSTR_FREE(dns_req->dnsreq_query);
	SM_STR_FREE(dns_req->dnsreq_key);
	if (dns_mgr_ctx != NULL) {
		SM_IS_DNSM_CTX(dns_mgr_ctx);
		if (DNSREQ_IS_ININCLIST(dns_req)) {
MTA_LIBDNS_DBG_DPRINTF((smioerr, "reqfree, dns_req=%p\n", dns_req));
			DNSIRQL_REMOVE(dns_mgr_ctx, dns_req);
		}
		else if (DNSREQ_IS_INWAITLIST(dns_req))
			DNSWRQL_REMOVE(dns_mgr_ctx, dns_req);
	}
	sm_free_size(dns_req, sizeof(*dns_req));
	return SM_SUCCESS;
}

/*
**  DNS_REQ_DEL -- free a DNS request
**	(wrapper for bht_*(); still necessary?)
**
**	Parameters:
**		value -- DNS request
**		key -- key (ignored)
**		ctx -- DNS manager context
**
**	Return value:
**		usual sm_error code
**
**	Locking: dns_mgr_ctx must be locked by caller.
*/

void
dns_req_del(void *value, void *key, void *ctx)
{
	MTA_LIBDNS_DBG_DPRINTF((smioerr, "sev=DBG, func=dns_req_del, dns_req=%p\n",
		value));
	(void) dns_req_free((dns_req_P) value, (dns_mgr_ctx_P) ctx);
}

/*
**  DNS_CRT_KEY -- create a key for a DNS query
**
**	Parameters:
**		query -- DNS query
**		type -- DNS type
**		pkey -- (pointer to) key (output)
**		pkeylen -- (pointer to) keylen (output)
**
**	Return value:
**		usual sm_error code
*/

static sm_ret_T
dns_crt_key(sm_cstr_P query, dns_type_T type, sm_str_P *pkey, size_t *pkeylen)
{
	sm_str_P key;
	size_t keylen;
	sm_ret_T ret;

	SM_REQUIRE(pkey != NULL);
	SM_REQUIRE(pkeylen != NULL);

	keylen = sm_cstr_getlen(query) + sizeof(type);
	key = sm_str_new(NULL, keylen, keylen + 4);
	if (NULL == key) {
		ret = sm_error_temp(SM_EM_DNS, ENOMEM);
		goto error;
	}
	ret = sm_str_scatn(key, sm_cstr_data(query), sm_cstr_getlen(query));
	if (sm_is_err(ret))
		goto error;
	ret = sm_str_rm_trail(key, ".");
	if (sm_is_err(ret))
		goto error;
	keylen -= ret;		/* decrement keylen if '.' has been removed */

	/* always lower case?! */
	sm_str2lower(key);
	ret = sm_str_scatn(key, (const uchar *) &type, sizeof(type));
	if (sm_is_err(ret))
		goto error;
	SM_ASSERT(keylen == sm_str_getlen(key));

	*pkey = key;
	*pkeylen = keylen;
	return SM_SUCCESS;

  error:
	if (key != NULL)
		sm_str_free(key);
	*pkey = NULL;
	*pkeylen = 0;
	return ret;
}

/*
**  DNS_TSK_SELECT -- select a DNS task
**
**	Parameters:
**		dns_mgr_ctx -- DNS manager context
**		now -- current time
**		wakeup -- "wake up" the selected DNS task?
**		pidx -- (pointer to) selected DNS task (output)
**		locktype -- kind of locking
**
**	Return value:
**		usual sm_error code
**
**	Locking: locks dns_mgr_ctx and dnsmgr_tsk if requested
*/

static sm_ret_T
dns_tsk_select(dns_mgr_ctx_P dns_mgr_ctx, time_T now, bool wakeup, ushort *pidx, thr_lock_T locktype)
{
	sm_ret_T ret;
	ushort s_tsk, j, i;
	time_T chg;

	SM_IS_DNSM_CTX(dns_mgr_ctx);
	C_THREAD_LOCK(thr_lock_it(locktype), &dns_mgr_ctx->dnsmgr_mutex);

	/* round robin?? or select server with least open queries? */
	chg = TIME_T_MAX;
	i = USHRT_MAX;
	s_tsk = dns_mgr_ctx->dnsmgr_ctsk;
	ret = SM_SUCCESS;
	for (j = 0; j < SM_DNS_TSKS_MAX; j++) {
		/* found a working DNS task? */
		if (dns_mgr_ctx->dnsmgr_tsk[s_tsk] != NULL &&
		    dns_mgr_ctx->dnsmgr_tskstatus[s_tsk] == DNSTSK_ST_OK)
			break;
#if 0
// 		if (dns_mgr_ctx->dnsmgr_tskchg[s_tsk] > 0 &&
// 		    dns_mgr_ctx->dnsmgr_tskstatus[s_tsk] == DNSTSK_ST_EOF)
// 		{
// 			ret = dns_reconnect(tsk);  tsk is not available...
// 		}
#endif
		if (dns_mgr_ctx->dnsmgr_tskchg[s_tsk] > 0 &&
		    dns_mgr_ctx->dnsmgr_tskstatus[s_tsk] == DNSTSK_ST_DEAD &&
		    dns_mgr_ctx->dnsmgr_tskchg[s_tsk] < chg)
		{
			chg = dns_mgr_ctx->dnsmgr_tskchg[s_tsk];
			i = s_tsk;
		}
		s_tsk = ++dns_mgr_ctx->dnsmgr_ctsk;
		if (s_tsk >= dns_mgr_ctx->dnsmgr_ntsks)
			s_tsk = dns_mgr_ctx->dnsmgr_ctsk = 0;
	}
	if (j >= SM_DNS_TSKS_MAX) {
		if (i < dns_mgr_ctx->dnsmgr_ntsks) {
			s_tsk = i;
			dns_mgr_ctx->dnsmgr_tskstatus[s_tsk] = DNSTSK_ST_OK;
			dns_mgr_ctx->dnsmgr_tskchg[s_tsk] = now;
			sm_log_write(dns_mgr_ctx->dnsmgr_lctx,
				LIBDNS_LCAT_RESOLVER, LIBDNS_LMOD_RESOLVER,
				SM_LOG_WARN, 4,
				"sev=WARN, func=dns_tsk_select, status=all dns servers considered unresponsive, action=selected one with oldest status change");
		}
		else if ((s_tsk = dns_mgr_ctx->dnsmgr_ctsk) < dns_mgr_ctx->dnsmgr_ntsks) {
			sm_log_write(dns_mgr_ctx->dnsmgr_lctx,
				LIBDNS_LCAT_RESOLVER, LIBDNS_LMOD_RESOLVER,
				SM_LOG_ERROR, 1,
				"sev=ERROR, func=dns_tsk_select, status=all dns servers considered unresponsive, action=selected one round robin");
		}
		else {
			sm_log_write(dns_mgr_ctx->dnsmgr_lctx,
				LIBDNS_LCAT_RESOLVER, LIBDNS_LMOD_RESOLVER,
				SM_LOG_FATAL, 1,
				"sev=FATAL, func=dns_tsk_select, status=no dns server available");
			ret = sm_error_perm(SM_EM_DNS, SM_E_NOTFOUND);
		}
	}
	MTA_LIBDNS_DBG_DPRINTF((smioerr, "sev=DBG, func=dns_tsk_select, j=%hu, i=%hu, wakeup=%d\n", j, i, wakeup));

	if (wakeup && s_tsk < dns_mgr_ctx->dnsmgr_ntsks) {
		dns_tsk_P dns_tsk;

		dns_tsk = dns_mgr_ctx->dnsmgr_dnstsks[s_tsk];
		SM_IS_DNSTSK(dns_tsk);
		C_THREAD_LOCK(thr_lock_it_s1(locktype), &dns_tsk->dnstsk_mutex);

		/* access is locked if locktype is set properly */
		if (!DNS_TSK_IS_FLAG(dns_tsk, DNS_TSK_FL_WR_EN)) {

			/*
			 *  Ignore return value: in the worst case the request will
			 *  not be sent to a DNS server but a timeout error will be
			 *  returned to the caller, hence we'll ignore any error
			 *  (for now).
			 */

			(void) evthr_en_wr(dns_mgr_ctx->dnsmgr_tsk[s_tsk]);
			DNS_TSK_SET_FLAG(dns_tsk, DNS_TSK_FL_WR_EN);
		}
		C_THREAD_UNLOCK(thr_unl_always_s1(locktype), &dns_tsk->dnstsk_mutex);
	}

	C_THREAD_UNLOCK(thr_unl_always(locktype), &dns_mgr_ctx->dnsmgr_mutex);
	if (pidx != NULL && ! sm_is_err(ret))
		*pidx = s_tsk;
	return ret;
}

/*
**  DNS_REQ_ADD -- Add a DNS request to the queue.
**	This adds a DNS request to the hash table and if it isn't yet in the
**	request list also to that.
**
**	Parameters:
**		dns_mgr_ctx -- DNS manager context
**		query -- DNS query (MUST be '\0' terminated; SM_CSTR_DUP()ed)
**		type -- DNS type
**		timeout -- timeout (if 0: use default)
**		fct -- callback to invoke with DNS result
**		ctx -- context to pass to callback
**
**	Return value:
**		usual sm_error code: an error will only be returned
**		if the request was NOT added to the request list and hence
**		the caller can't expect a reply.
**
**	Locking: dns_mgr_ctx is locked as necessary.
*/

sm_ret_T
dns_req_add(dns_mgr_ctx_P dns_mgr_ctx, sm_cstr_P query, dns_type_T type, uint timeout, dns_callback_F *fct, void *ctx)
{
	sm_ret_T ret;
	bool reqnotinht;
	dns_req_P dns_req;
	bht_entry_P bhte;
	sm_str_P key;
	size_t keylen;
	time_T now;
	timeval_T nowt;

	SM_IS_DNSM_CTX(dns_mgr_ctx);
	SM_REQUIRE(query != NULL);
	dns_req = (dns_req_P) sm_zalloc(sizeof(*dns_req));
	if (NULL == dns_req)
		return sm_error_temp(SM_EM_DNS, ENOMEM);
	ret = SM_SUCCESS;
	key = NULL;
	reqnotinht = false;
	dns_req->dnsreq_query = SM_CSTR_DUP(query);
	dns_req->dnsreq_type = type;

	(void) evthr_timeval(dns_mgr_ctx->dnsmgr_cleanup->evthr_t_ctx, &nowt);
#if DNS_STATISTICS
	dns_req->dnsreq_startt = nowt;
#endif

	now = nowt.tv_sec;
	dns_req->dnsreq_endt = now +
			((timeout == 0) ? dns_mgr_ctx->dnsmgr_timeout : timeout);

	MTA_LIBDNS_DBG_DPRINTF((smioerr,
		"sev=DBG, func=dns_req_add, now=%ld, endt=%ld\n",
		(long) now, (long) dns_req->dnsreq_endt));

	/*
	**  Create key from query and type... this is still ugly...
	**  Note: we can't use the output of res_mkquery...
	*/

	ret = dns_crt_key(query, type, &key, &keylen);
	if (sm_is_err(ret))
		goto error;
	dns_req->dnsreq_key = key;
	dns_req->dnsreq_fct = fct;
	dns_req->dnsreq_ctx = ctx;
	THREAD_LOCK(&dns_mgr_ctx->dnsmgr_mutex);

	/* Is there already a request for this? */
	bhte = bht_locate(dns_mgr_ctx->dnsmgr_req_ht,
			(const char *) sm_str_data(key), (uint) sm_str_getlen(key));

	reqnotinht = (NULL == bhte);

	/* Always add the request to the hash table so it can be answered */
	ret = bht_add(dns_mgr_ctx->dnsmgr_req_ht,
			(char *) sm_str_data(dns_req->dnsreq_key),
			(uint) sm_str_getlen(dns_req->dnsreq_key), dns_req, &bhte);
#if LIBDNS_PARANOIA
	if (!sm_is_err(ret) && !reqnotinht) {
		dns_req_P dr, drp;
		bool found;

		found = false;
		drp = NULL;
		dr = DNSWRQL_FIRST(dns_mgr_ctx);
		while (dr != DNSWRQL_END(dns_mgr_ctx)) {
			SM_ASSERT(dr != drp);
			found = sm_str_eq(dns_req->dnsreq_key, dr->dnsreq_key);
			if (found)
				break;
			drp = dr;
			dr = DNSWRQL_NEXT(dr);
		}
		if (!found) {
			drp = NULL;
			dr = DNSIRQL_FIRST(dns_mgr_ctx);
			while (dr != DNSIRQL_END(dns_mgr_ctx)) {
				SM_ASSERT(dr != drp);
				found = sm_str_eq(dns_req->dnsreq_key, dr->dnsreq_key);
				if (found) {
					/* SM_ASSERT(DNS_TSK_IS_FLAG(dns_tsk, DNS_TSK_FL_WR_EN)); */
					break;
				}
				drp = dr;
				dr = DNSWRQL_NEXT(dr);
			}
		}
		SM_ASSERT(found);
	}
#endif

	if (sm_is_err(ret)) {
		sm_log_write(dns_mgr_ctx->dnsmgr_lctx,
			LIBDNS_LCAT_RESOLVER, LIBDNS_LMOD_RESOLVER,
			SM_LOG_ERROR, 4,
			"sev=ERROR, func=dns_req_add, bhte=%p, bht_add=%r, key='%s'[%p], len=%d\n",
			bhte, ret, sm_str_data(dns_req->dnsreq_key), dns_req->dnsreq_key,
			sm_str_getlen(dns_req->dnsreq_key));
	}
	else if (reqnotinht) {
		/* entry was not in hash table: add it to the list */
		DNSIRQL_INSERT_TAIL(dns_mgr_ctx, dns_req);
		DNSREQ_SET_INLIST(dns_req, DNSREQ_FL_IN_INC);
	}

	MTA_LIBDNS_DBG_DPRINTF((smioerr,
		"sev=DBG, func=dns_req_add, bht_add bhte=%p, ret=%x, key='%s'[%p], len=%d, notinht=%d, empty=%d, dns_req=%p\n",
		bhte, ret, sm_str_data(dns_req->dnsreq_key), dns_req->dnsreq_key,
		sm_str_getlen(dns_req->dnsreq_key), reqnotinht, DNSIRQL_EMPTY(dns_mgr_ctx), dns_req));
		

#if MTA_USE_PTHREADS
	THREAD_UNLOCK(&dns_mgr_ctx->dnsmgr_mutex);

	if (sm_is_success(ret) && reqnotinht) {
		(void) dns_tsk_select(dns_mgr_ctx, now, true, NULL,
				 THR_LOCK_UNLOCK|THR_LOCK_UNLOCK_S1);

		/*
		 *  Currently the cleanup task should only be activated if the request
		 *  has been added to waitq because the cleanup task reads that queue.
		 */

		if (sm_is_success(ret) && dns_mgr_ctx->dnsmgr_cleanup != NULL) {
			timeval_T slpt, ct, newt;

			/* set timeout for cleanup task */
			(void) evthr_timeval(dns_mgr_ctx->dnsmgr_cleanup->evthr_t_ctx, &ct);
			slpt.tv_usec = 0;
			slpt.tv_sec = dns_mgr_ctx->dnsmgr_timeout;
			timeradd(&ct, &slpt, &newt);
			(void) evthr_new_sl(dns_mgr_ctx->dnsmgr_cleanup, newt, false);
		}
	}
#endif /* MTA_USE_PTHREADS */

	return ret;

  error:
	if (dns_req != NULL)
		dns_req_free(dns_req, NULL);
	return ret;
}

/* context for dns_bht2reql_fct() */
struct bht2reql_S
{
	dns_mgr_ctx_P	 b2r_dns_mgr_ctx;
	dns_rql_P	 b2r_rql_hd;
};

typedef struct bht2reql_S	bht2reql_T, *bht2reql_P;

/*
**  DNS_BHT2REQL_FCT -- move DNS requests from hash table into local list.
**	callback for bht_rm_all() used in dns_bht2reql() below.
**
**	Parameters:
**		value -- DNS request
**		key -- key (ignored, for compatibility with bhfree_F)
**		ctx -- context (see above: bht2reql_P)
**
**	Return value:
**		usual sm_error code
**
**	Locking: dns_mgr_ctx must be locked by caller.
*/

static void
dns_bht2reql_fct(void *value, void *key, void *ctx)
{
	dns_req_P dns_req;
	bht2reql_P bht2reql;
	dns_mgr_ctx_P dns_mgr_ctx;
	dns_rql_P rql_hd;

	dns_req = (dns_req_P) value;
	bht2reql = (bht2reql_P) ctx;
	if (NULL == dns_req || bht2reql == NULL) {
		/* abort for now */
		SM_ASSERT(dns_req != NULL);
		SM_ASSERT(bht2reql != NULL);
		MTA_LIBDNS_DBG_DPRINTF((smioerr,
			"sev=ERROR, func=dns_bht2reql_fct, dns_req=%p, bht2reql=%p\n",
			dns_req, bht2reql));
		return;
	}
	dns_mgr_ctx = bht2reql->b2r_dns_mgr_ctx;
	SM_IS_DNSM_CTX(dns_mgr_ctx);
	rql_hd = bht2reql->b2r_rql_hd;
	if (NULL == dns_mgr_ctx || rql_hd == NULL) {
		/* abort for now */
		SM_ASSERT(dns_mgr_ctx != NULL);
		SM_ASSERT(rql_hd != NULL);
		MTA_LIBDNS_DBG_DPRINTF((smioerr,
			"sev=ERROR, func=dns_bht2reql_fct, dns_mgr_ctx=%p, rql_hd=%p\n",
			dns_mgr_ctx, rql_hd));
		return;
	}

	MTA_LIBDNS_DBG_DPRINTF((smioerr,
		"sev=DBG, func=dns_bht2reql_fct, dns_req=%p, inlist=%d\n",
		dns_req, DNSREQ_IS_INANYLIST(dns_req)));

	/* Remove entry from dns_mgr list (if it is in that list) */
	if (DNSREQ_IS_ININCLIST(dns_req)) {
MTA_LIBDNS_DBG_DPRINTF((smioerr, "ht2r, dns_req=%p\n", dns_req));
		DNSIRQL_REMOVE(dns_mgr_ctx, dns_req);
		DNSREQ_CLR_INLIST(dns_req);
	}
	else if (DNSREQ_IS_INWAITLIST(dns_req)) {	/* possible?? */
		DNSWRQL_REMOVE(dns_mgr_ctx, dns_req);
		DNSREQ_CLR_INLIST(dns_req);
	}

	/* Add to local list */
	RQL_INSERT_TAIL(rql_hd, dns_req);
}

/*
**  DNS_BHT2REQL -- move DNS requests that match key from hash table
**		into local list.
**
**	Parameters:
**		dns_mgr_ctx -- DNS manager context
**		key -- key to match
**		keylen -- length of key
**		rql_hd -- head of DNS request list to which items are moved
**
**	Return value:
**		usual sm_error code
**
**	Locking: dns_mgr_ctx must be locked by caller.
*/

static sm_ret_T
dns_bht2reql(dns_mgr_ctx_P dns_mgr_ctx, sm_str_P key, size_t keylen, dns_rql_P rql_hd)
{
	bht2reql_T bht2reql;

	SM_IS_DNSM_CTX(dns_mgr_ctx);
	SM_ASSERT(rql_hd != NULL);
	SM_ASSERT(key != NULL);

	bht2reql.b2r_dns_mgr_ctx = dns_mgr_ctx;
	bht2reql.b2r_rql_hd = rql_hd;

	/*
	**  Remove all entries that match key/keylen from hash table;
	**  remove them from DNS manager list, add to local list.
	**  Use the bht_rm_all() function callback to achieve this
	**  instead of accessing the data directly.
	*/

	bht_rm_all(dns_mgr_ctx->dnsmgr_req_ht,
		(const char *) sm_str_data(key), keylen, dns_bht2reql_fct, &bht2reql);

	return SM_SUCCESS;
}

/*
**  DNS_TSK_RD -- DNS communication task: read replies from DNS server
**
**	Parameters:
**		tsk -- evthr task
**
**	Return value:
**		usual sm_error code
**
**	Locking: dns_mgr_ctx is locked when necessary.
*/

static sm_ret_T
dns_tsk_rd(sm_evthr_task_P tsk)
{
	sm_ret_T ret, res;
	dns_tsk_P dns_tsk;
	dns_res_P dns_res;
	dns_req_P dns_req, dns_req_next;
	dns_mgr_ctx_P dns_mgr_ctx;
	bht_entry_P bhte;
	uchar query[MAXHOSTNAMELEN];
	ushort type;
	sm_str_P key;
	size_t keylen;
	dns_rql_T rql_hd;

#if MTA_USE_PTHREADS
	SM_IS_EVTHR_TSK(tsk);
#endif
	dns_tsk = (dns_tsk_P) tsk->evthr_t_actx;
	if (NULL == dns_tsk)
		return EVTHR_DEL;
	if (!is_valid_fd(tsk->evthr_t_fd))
		return EVTHR_DEL;
	SM_IS_DNSTSK(dns_tsk);
	key = NULL;
	res = dns_receive(dns_tsk);
	MTA_LIBDNS_DBG_DPRINTF((smioerr,
		"sev=DBG, func=dns_tsk_rd, receive=%x\n", res));
	if (sm_is_err(res)) {
		/* XXX deal with it... */;
		ret = res;
		goto error;
	}
	if (res == 0)
		goto done;

	ret = dns_res_new(MAXRESHOSTS, &dns_res);
	if (sm_is_err(ret)) {
		/* XXX deal with it... */;
		goto error;
	}

	ret = dns_decode(dns_tsk->dnstsk_rd, query, sizeof(query), &type, dns_res);
	MTA_LIBDNS_DBG_DPRINTF((smioerr,
		"sev=DBG, func=dns_tsk_rd, decode=%x, query=%s\n", ret, query));

#if MTA_LIBDNS_TEST
	/* introduce some errors for some requests; ToDo: add others! */
#define SMDT_ARPA_TEMP	"190.0.0.127.in-addr.arpa"
#define SMDT_ARPA_PERM	"191.0.0.127.in-addr.arpa"
#define SMDT_ARPA_EINVAL	"192.0.0.127.in-addr.arpa"
#define SMDT_MX_EINVAL	"mxinvalid.sm9.org"
	if (sm_memeq(query, SMDT_ARPA_TEMP, sizeof(SMDT_ARPA_TEMP) - 1))
		dns_res->dnsres_ret = DNSR_TEMP;
	else if (sm_memeq(query, SMDT_ARPA_PERM, sizeof(SMDT_ARPA_PERM) - 1))
		dns_res->dnsres_ret = DNSR_PERM;
	else if (sm_memeq(query, SMDT_ARPA_EINVAL, sizeof(SMDT_ARPA_EINVAL) - 1))
		dns_res->dnsres_ret = sm_error_perm(SM_EM_DNS, EINVAL);
	else if (sm_memeq(query, SMDT_MX_EINVAL, sizeof(SMDT_MX_EINVAL) - 1))
		dns_res->dnsres_ret = DNSR_MXINVALID;
#endif /* MTA_LIBDNS_TEST */

	dns_mgr_ctx = dns_tsk->dnstsk_mgr;
	SM_IS_DNSM_CTX(dns_mgr_ctx);

	/* Create key from query and type... this is still ugly... */
	keylen = strlen((char *) query) + sizeof(type);
	key = sm_str_new(NULL, keylen, keylen + 4);
	if (NULL == key) {
		ret = sm_error_temp(SM_EM_DNS, ENOMEM);
		goto error;
	}
	ret = sm_str_scatn(key, query, strlen((char *) query));
	if (sm_is_err(ret))
		goto error;
	ret = sm_str_rm_trail(key, ".");
	if (sm_is_err(ret))
		goto error;
	keylen -= ret;		/* decrement keylen if '.' has been removed */

	/* always lower case?! */
	sm_str2lower(key);
	ret = sm_str_scatn(key, (const uchar *) &type, sizeof(type));
	if (sm_is_err(ret))
		goto error;
	SM_ASSERT(keylen == sm_str_getlen(key));
	RQL_INIT(&rql_hd);

	THREAD_LOCK(&dns_mgr_ctx->dnsmgr_mutex);

	/*
	**  A two step approach is required here:
	**  1. Build a new list of requests (removing them from dns_mgr_ctx:
	**	request list and hash table)
	**  2. Use that list to invoke the callbacks
	**  This is required to get around locking/concurrent access issues:
	**  the table/list can't be locked when invoking the callback
	**  since otherwise the callback can't invoke a function like
	**  dns_req_add() which locks the dns_mgr_ctx table/list.
	**
	**  First lookup the entry and check whether the request id
	**  matches the result id. If it doesn't: ignore it.
	**  Note: bhtable prepends new entries, hence it is necessary to
	**  walk to the end of the chain. This is ugly and slow; a better
	**  method should be used!
	**  NOTE: this depends on the implementation of bhtable!
	**  It should use some bht function instead!
	*/

	bhte = bht_locate(dns_mgr_ctx->dnsmgr_req_ht,
			(const char *) sm_str_data(key), keylen);

	dns_req = NULL;
	while (bhte != NULL) {
		dns_req = (dns_req_P) (bhte->bhe_value);
		SM_ASSERT(dns_req != NULL);
		if (dns_req->dnsreq_id == dns_res->dnsres_id)
			break;
		bhte = bhte->bhe_next;
#if 0
		if (dns_req->dnsreq_id != 0) {
sm_io_fprintf(smioerr, "sev=WARN, func=dns_tsk_rd, dns_req=%p, dns_req_id=%d, dnsres_id=%d, status=id_mismatch, where=in_while\n",
dns_req, NULL == dns_req ? -1 : dns_req->dnsreq_id, dns_res->dnsres_id);
		}
#endif
	}

	if (NULL == bhte || dns_req == NULL) {
		THREAD_UNLOCK(&dns_mgr_ctx->dnsmgr_mutex);
		if (dns_req != NULL) {
			sm_log_write(dns_mgr_ctx->dnsmgr_lctx,
				LIBDNS_LCAT_RESOLVER, LIBDNS_LMOD_RESOLVER,
				SM_LOG_WARN, 8,
				"sev=WARN, func=dns_tsk_rd, dns_req=%p, dns_req_id=%d, dnsres_id=%d, status=id_mismatch/not_found, key=%#N",
				dns_req, NULL == dns_req ? -1 : dns_req->dnsreq_id,
				dns_res->dnsres_id, key);
		}
		goto freeit;
	}

	ret = dns_bht2reql(dns_mgr_ctx, key, keylen, &rql_hd);
	/* result ignored for now; it's always SUCCESS */

	THREAD_UNLOCK(&dns_mgr_ctx->dnsmgr_mutex);

	/* Go through list of "done" requests and invoke callbacks */
	for (dns_req = RQL_FIRST(&rql_hd);
	     dns_req != RQL_END(&rql_hd);
	     dns_req = dns_req_next)
	{
		MTA_LIBDNS_DBG_DPRINTF((smioerr,
			"sev=DBG, func=dns_tsk_rd, dns_req=%p, inlist=%d\n",
			dns_req, DNSREQ_IS_INANYLIST(dns_req)));
		dns_req_next = RQL_NEXT(dns_req);
		if (dns_req->dnsreq_id != 0 && dns_req->dnsreq_id != dns_res->dnsres_id)
		{
			sm_log_write(dns_mgr_ctx->dnsmgr_lctx,
				LIBDNS_LCAT_RESOLVER, LIBDNS_LMOD_RESOLVER,
				SM_LOG_WARN, 7,
				"sev=WARN, func=dns_tsk_rd, dns_req_id=%d, dnsrpl_id=%d\n",
				dns_req->dnsreq_id, dns_res->dnsres_id);
		}

		if (dns_req->dnsreq_fct != NULL) {
			dns_res->dnsres_query = dns_req->dnsreq_query;

			/* Ignore return value? */
			dns_req->dnsreq_fct(dns_res, dns_req->dnsreq_ctx);
		}

		/* Unlink entry from list */
		RQL_REMOVE(&rql_hd, dns_req);
		ret = dns_req_free(dns_req, NULL);
	}

  freeit:
	/* Free dns_res */
	ret = dns_res_free(dns_res);
	sm_str_free(key);

  done:
	return EVTHR_WAITQ;

  error:
	MTA_LIBDNS_DBG_DPRINTF((smioerr, "sev=ERROR, func=dns_tsk_rd, ret=%x\n",
		ret));
	if (key != NULL)
		sm_str_free(key);
	return ret;
}

/*
**  DNS_REQ_INS_WRQL -- insert DNS request into wait list
**
**	Parameters:
**		dns_mgr_ctx -- DNS manager context
**		dns_req -- DNS request
**
**	Return value:
**		SM_SUCCESS
**
**	Locking: dns_mgr_ctx must be locked by caller.
*/

sm_ret_T
dns_req_ins_wrql(dns_mgr_ctx_P dns_mgr_ctx, dns_req_P dns_req)
{
	dns_req_P dns_reqh;

	SM_IS_DNSM_CTX(dns_mgr_ctx);
	if (DNSWRQL_EMPTY(dns_mgr_ctx)) {
		DNSWRQL_INSERT_TAIL(dns_mgr_ctx, dns_req);
		MTA_LIBDNS_DBG_DPRINTF((smioerr,
			"sev=DBG, func=dns_tsk_wr, writelist=empty\n"));
	}
	else {
		/*
		**  Insert dns_req at the right place in the list:
		**  start from the back and go to the front until
		**  an entry is reached whose "end time" is not
		**  greater than the one of the entry to be inserted.
		*/

		dns_reqh = DNSWRQL_LAST(dns_mgr_ctx);
		while (dns_reqh != DNSWRQL_FIRST(dns_mgr_ctx) &&
		       dns_req->dnsreq_endt < dns_reqh->dnsreq_endt)
		{
			dns_reqh = DNSWRQL_PREV(dns_reqh);
		}
		MTA_LIBDNS_DBG_DPRINTF((smioerr,
			"sev=DBG, func=dns_tsk_wr, req=%p, reqh=%p, req->endt=%ld, reqh->ednt=%ld\n",
			dns_req, dns_reqh, (long) dns_req->dnsreq_endt,
			(long) dns_reqh->dnsreq_endt));
		if (dns_req->dnsreq_endt >= dns_reqh->dnsreq_endt)
			DNSWRQL_INSERT_AFTER(dns_mgr_ctx, dns_reqh, dns_req);
		else
			DNSWRQL_INSERT_BEFORE(dns_mgr_ctx, dns_reqh, dns_req);
	}
	DNSREQ_SET_INLIST(dns_req, DNSREQ_FL_IN_WAIT);
	return SM_SUCCESS;
}

/*
**  DNS_TSK_WR -- DNS communication task: send queries to DNS server
**
**	Parameters:
**		tsk -- evthr task
**
**	Return value:
**		usual sm_error code
*/

static sm_ret_T
dns_tsk_wr(sm_evthr_task_P tsk)
{
	sm_ret_T ret;
	int r;
	dns_tsk_P dns_tsk;
	dns_mgr_ctx_P dns_mgr_ctx;
	dns_req_P dns_req;
	bool empty;

	SM_IS_EVTHR_TSK(tsk);
	dns_tsk = (dns_tsk_P) tsk->evthr_t_actx;
	if (NULL == dns_tsk)
		return EVTHR_DEL;
	if (!is_valid_fd(tsk->evthr_t_fd))
		return EVTHR_DEL;
	SM_IS_DNSTSK(dns_tsk);
	dns_mgr_ctx = dns_tsk->dnstsk_mgr;
	SM_IS_DNSM_CTX(dns_mgr_ctx);
	dns_req = NULL;

	do {
		THREAD_LOCK(&dns_mgr_ctx->dnsmgr_mutex);

		empty = DNSIRQL_EMPTY(dns_mgr_ctx);
		MTA_LIBDNS_DBG_DPRINTF((smioerr,
				"sev=DBG, func=dns_tsk_wr, srv=%A, empty=%d\n",
						dns_tsk->dnstsk_sin.sin_addr.s_addr, empty));
		if (!empty) {
			/* move element from incoming to wait list */
			dns_req = DNSIRQL_FIRST(dns_mgr_ctx);
MTA_LIBDNS_DBG_DPRINTF((smioerr, "tsk_wr, dns_req=%p\n", dns_req));
			DNSIRQL_REMOVE(dns_mgr_ctx, dns_req);
			dns_req->dnsreq_idx = dns_tsk->dnstsk_idx;

			(void) dns_req_ins_wrql(dns_mgr_ctx, dns_req);
			THREAD_UNLOCK(&dns_mgr_ctx->dnsmgr_mutex);
		}
		else {
#if MTA_USE_PTHREADS
			THREAD_LOCK(&dns_tsk->dnstsk_mutex);
			DNS_TSK_CLR_FLAG(dns_tsk, DNS_TSK_FL_WR_EN);
			THREAD_UNLOCK(&dns_tsk->dnstsk_mutex);
			THREAD_UNLOCK(&dns_mgr_ctx->dnsmgr_mutex);
#endif /* MTA_USE_PTHREADS */

			/* Turn off WR, dns_req_add() will turn it on if required. */
			return EVTHR_WAITQ|evthr_r_no(EVTHR_EV_WR);
		}

		MTA_LIBDNS_DBG_DPRINTF((smioerr, "sev=DBG, func=dns_wr, srv=%A, req=%p\n",
			dns_tsk->dnstsk_sin.sin_addr.s_addr, dns_req));

		/* Extract query from dns_req and send it to DNS server */
		do {
			/* Note: dnsreq_query must be '\0' terminated! */
#if MTA_USE_DNSSEC
			r = ns_nmkquery((const char *) sm_cstr_data(dns_req->dnsreq_query),
					C_IN, dns_req->dnsreq_type,
					sm_str_data(dns_tsk->dnstsk_wr), (int) sm_str_getsize(dns_tsk->dnstsk_wr));
#else
			r = res_mkquery(QUERY,
				(const char *) sm_cstr_data(dns_req->dnsreq_query),
				C_IN, dns_req->dnsreq_type, NULL, 0,
				NULL, sm_str_data(dns_tsk->dnstsk_wr),
				(int) sm_str_getsize(dns_tsk->dnstsk_wr));
#endif

			if (r == -1) {
				size_t len;

				len = sm_str_getsize(dns_tsk->dnstsk_wr) * 2;
				ret = sm_str_space(dns_tsk->dnstsk_wr, len);
				if (sm_is_err(ret)) {
					sm_log_write(dns_mgr_ctx->dnsmgr_lctx,
						LIBDNS_LCAT_RESOLVER, LIBDNS_LMOD_RESOLVER,
						SM_LOG_ERROR, 4,
						"sev=ERROR, func=dns_tsk_wr, srv=%A, increase_buffer=%m",
						dns_tsk->dnstsk_sin.sin_addr.s_addr, ret);
					goto error;
				}
			}
			else {
				HEADER *hp;

#if MTA_USE_DNSSEC && defined(RES_USE_EDNS0) && MTA_USE_EDNS0
				r = ns_nopt(r, sm_str_data(dns_tsk->dnstsk_wr), (int) sm_str_getsize(dns_tsk->dnstsk_wr), 4096);
#endif
				hp = (HEADER *) sm_str_data(dns_tsk->dnstsk_wr);
				dns_req->dnsreq_id = hp->id;
			}
		} while (r == -1);
		SM_ASSERT(r >= 0);
		SM_STR_SETLEN(dns_tsk->dnstsk_wr, (uint)r);

		ret = dns_send(dns_tsk);
		MTA_LIBDNS_DBG_DPRINTF((smioerr,
			"sev=DBG, func=dns_tsk_wr, dns_send=%#x\n", ret));
		if (sm_is_err(ret)) {
			sm_log_write(dns_mgr_ctx->dnsmgr_lctx,
				LIBDNS_LCAT_RESOLVER, LIBDNS_LMOD_RESOLVER,
				SM_LOG_ERROR, 4,
				"sev=ERROR, func=dns_tsk_wr, srv=%A, dns_send=%m",
					dns_tsk->dnstsk_sin.sin_addr.s_addr, ret);
			
		}

	} while (sm_is_success(ret) && 0 == sm_write_wait(dns_tsk->dnstsk_fd, 0));
#if 0
	/*
	**  Don't terminate the task just because dns_send() failed
	**  How to deal with this properly? (see below)
	*/

	if (sm_is_err(ret))
		goto error;
#endif

	return EVTHR_WAITQ;

  error:
	/*
	**  XXX need to determine whether this is "real" error,
	**	i.e., one that actually will persist and hence it makes
	**	sense to terminate this task (return EVTHR_DEL).
	*/

	return EVTHR_WAITQ;
}

#if MTA_LIBDNS_DEBUG
sm_ret_T
printreq(bht_entry_P e, void *ctx)
{
	if (e != NULL)
		sm_io_fprintf(smioerr, "e=%p, ctx=%p\n", e, ctx);
	return 0;
}
#endif

/*
**  DNS_TSK_NOTIFY -- deal with requests that have not been answered
**
**	Parameters:
**		dns_mgr_ctx -- DNS manager context
**		all --  time out all requests (e.g., on shutdown)?
**		now --  current time
**		sleept --  for statistics
**		pdelay --  (pointer to) delay to next run (output, can be NULL)
**
**	Return value:
**		usual sm_error code
*/

sm_ret_T
dns_tsk_notify(dns_mgr_ctx_P dns_mgr_ctx, bool all, time_T now, timeval_T sleept, uint *pdelay)
{
	sm_ret_T ret;
	dns_req_P dns_req, dns_req_next;
	dns_res_T dns_res;
	uint delay;
	sm_str_P key;
	size_t keylen;
	dns_rql_T rql_hd;

	ret = delay = 0;
	RQL_INIT(&rql_hd);
	dns_req_next = NULL;

	SM_IS_DNSM_CTX(dns_mgr_ctx);
	THREAD_LOCK(&dns_mgr_ctx->dnsmgr_mutex);

	while (!DNSWRQL_EMPTY(dns_mgr_ctx)) {
		dns_req = DNSWRQL_FIRST(dns_mgr_ctx);
		if (dns_req == dns_req_next) {
			sm_log_write(dns_mgr_ctx->dnsmgr_lctx,
				LIBDNS_LCAT_RESOLVER, LIBDNS_LMOD_RESOLVER, SM_LOG_ERROR, 4,
				"sev=ERROR, func=dns_tsk_notify, status=loop, dns_req=%p, dns_req_next=%p\n",
				dns_req, dns_req_next);
			SM_ASSERT(dns_req != dns_req_next);
			break;
		}
		dns_req_next = dns_req;

		MTA_LIBDNS_DBG_DPRINTF((smioerr,
			"sev=DBG, func=dns_tsk_notify, dns_req=%p, dns_req_next=%p\n",
			dns_req, dns_req_next));
		MTA_LIBDNS_DBG_DPRINTF((smioerr,
			"sev=DBG, func=dns_tsk_notify, now=%ld, endt=%ld\n",
			(long) now, (long) dns_req->dnsreq_endt));

		/* Stop the loop if the timeout hasn't been reached yet */
		if (now <= dns_req->dnsreq_endt && !all) {
			delay = dns_req->dnsreq_endt - now;
			if (delay == 0)
				delay = TSK_CLEANUP_SLP_MIN;
			break;
		}

		if (dns_req->dnsreq_tries < dns_mgr_ctx->dnsmgr_retries && !all) {
#if DNS_STATISTICS
			dns_req->dnsreq_startt = sleept;
#endif
			dns_req->dnsreq_endt = now + dns_mgr_ctx->dnsmgr_timeout;
			DNSWRQL_REMOVE(dns_mgr_ctx, dns_req);
			DNSREQ_CLR_INLIST(dns_req);
			DNSIRQL_INSERT_TAIL(dns_mgr_ctx, dns_req);
			DNSREQ_SET_INLIST(dns_req, DNSREQ_FL_IN_INC);
			++dns_req->dnsreq_tries;

			/* how to avoid repeated activations? */
			(void) dns_tsk_select(dns_mgr_ctx, now, true, NULL,
				 THR_NO_LOCK|THR_LOCK_UNLOCK_S1);
			continue;
		}

		/* Create key from query and type... still ugly... */
		ret = dns_crt_key(dns_req->dnsreq_query, dns_req->dnsreq_type,
				&key, &keylen);
		if (sm_is_err(ret))
			goto errunl;

		MTA_LIBDNS_DBG_DPRINTF((smioerr,
			"sev=DBG, func=dns_tsk_notify, dns_req=%p, key='%s', keylen=%d\n",
			dns_req, sm_str_data(key), keylen));

		/*
		**  Remove dns_req from hash table.
		**  Notice: this removes all entries that match the key
		**  and moves them to our local list.
		**
		**  Note: this creates a problem: should we requeue
		**  those requests? This is just a timeout,
		**  maybe later on it actually works.
		*/

		ret = dns_bht2reql(dns_mgr_ctx, key, keylen, &rql_hd);
#if MTA_LIBDNS_DEBUG
		if (dns_req == DNSWRQL_FIRST(dns_mgr_ctx))
			bht_walk(dns_mgr_ctx->dnsmgr_req_ht, printreq, key);
#endif
		SM_ASSERT(dns_req != DNSWRQL_FIRST(dns_mgr_ctx));
		sm_str_free(key);
	}

	MTA_LIBDNS_DBG_DPRINTF((smioerr,
		"sev=DBG, func=dns_tsk_notify, empty=%d, delay=%d\n",
		DNSWRQL_EMPTY(dns_mgr_ctx), delay));
	if (!all && 0 == delay && !DNSWRQL_EMPTY(dns_mgr_ctx)) {
		dns_req = DNSWRQL_FIRST(dns_mgr_ctx);

		/* make sure the result will be > 0 [unsigned arithmetic] */
		if (now < dns_req->dnsreq_endt)
			delay = dns_req->dnsreq_endt - now;
		if (delay == 0)
			delay = TSK_CLEANUP_SLP_MIN;

		MTA_LIBDNS_DBG_DPRINTF((smioerr,
			"sev=DBG, func=dns_tsk_notify, where=not-empty, now=%ld, endt=%ld\n",
			(long) now, (long) dns_req->dnsreq_endt));
	}
	MTA_LIBDNS_DBG_DPRINTF((smioerr,
		"sev=DBG, func=dns_tsk_notify, where=end, delay=%d\n", delay));
	if (0 == delay)
		delay = TSK_CLEANUP_SLP;

	THREAD_UNLOCK(&dns_mgr_ctx->dnsmgr_mutex);

	/* Go through list of "done" requests and invoke callbacks */
	for (dns_req = RQL_FIRST(&rql_hd);
	     dns_req != RQL_END(&rql_hd);
	     dns_req = dns_req_next)
	{
		ushort u;
		dns_tsk_P dns_tsk;

		dns_req_next = RQL_NEXT(dns_req);
		MTA_LIBDNS_DBG_DPRINTF((smioerr,
			"sev=DBG, func=dns_tsk_notify, dns_req=%p, fct=%p\n",
			dns_req, dns_req->dnsreq_fct));

		/* mark "dead" DNS servers */
		if (!all && (u = dns_req->dnsreq_idx) < dns_mgr_ctx->dnsmgr_ntsks &&
		    (dns_tsk = dns_mgr_ctx->dnsmgr_dnstsks[u]) != NULL &&
		    dns_tsk->dnstsk_maxtimeouts > 0 &&
		    (++dns_tsk->dnstsk_timeouts > dns_tsk->dnstsk_maxtimeouts))
		{
			dns_mgr_ctx->dnsmgr_tskstatus[u] = DNSTSK_ST_DEAD;
			dns_mgr_ctx->dnsmgr_tskchg[u] = now;
			dns_tsk->dnstsk_timeouts = 0;
		}

		/* Unlink entry from list */
		if (dns_req->dnsreq_fct != NULL) {
			/* Return a timeout to caller... fill in more data? */
			sm_memzero(&dns_res, sizeof(dns_res));
			dns_res.dnsres_query = dns_req->dnsreq_query;
			dns_res.dnsres_ret = DNSR_TIMEOUT;
			dns_res.dnsres_qtype = dns_req->dnsreq_type;
			DRESL_INIT(&dns_res);

			/* Ignore return value? */
			dns_req->dnsreq_fct(&dns_res, dns_req->dnsreq_ctx);
		}
		RQL_REMOVE(&rql_hd, dns_req);
		ret = dns_req_free(dns_req, NULL);
	}

	if (pdelay != NULL)
		*pdelay = delay;
	return SM_SUCCESS;

  errunl:
	THREAD_UNLOCK(&dns_mgr_ctx->dnsmgr_mutex);
	if (pdelay != NULL)
		*pdelay = delay;
	return ret;
}

/*
**  DNS_TSK_CLEANUP -- deal with requests that have not been answered.
**	This should be run whenever a request times out.
**
**	Parameters:
**		tsk -- evthr task
**
**	Return value:
**		usual sm_error code
*/

static sm_ret_T
dns_tsk_cleanup(sm_evthr_task_P tsk)
{
	sm_ret_T ret;
	dns_mgr_ctx_P dns_mgr_ctx;
	time_T now;
	timeval_T sleept;
	uint delay;

	SM_IS_EVTHR_TSK(tsk);
	dns_mgr_ctx = (dns_mgr_ctx_P) tsk->evthr_t_actx;
	if (NULL == dns_mgr_ctx)
		return EVTHR_DEL;
	SM_IS_DNSM_CTX(dns_mgr_ctx);
	ret = evthr_timeval(tsk->evthr_t_ctx, &sleept);
	now = sleept.tv_sec;

	tsk->evthr_t_sleep.tv_usec = sleept.tv_usec;
	ret = dns_tsk_notify(dns_mgr_ctx, false, now, sleept, &delay);
	if (sm_is_err(ret))
		goto error;

	tsk->evthr_t_sleep.tv_sec = sleept.tv_sec + delay;
	return EVTHR_SLPQ;

  error:
	/* Cleanup? */
	tsk->evthr_t_sleep.tv_sec = sleept.tv_sec + delay;
	return EVTHR_SLPQ;	/* XXX Always? */
}

/*
**  DNS_DISCONNECT --  disconnect from a DNS server after getting EOF
**
**	Parameters:
**		tsk -- evthr task
**
**	Return value:
**		usual sm_error code
*/

static sm_ret_T
dns_disconnect(sm_evthr_task_P tsk)
{
	dns_tsk_P dns_tsk;

	SM_IS_EVTHR_TSK(tsk);
	dns_tsk = (dns_tsk_P) tsk->evthr_t_actx;
	SM_IS_DNSTSK(dns_tsk);
	if (is_valid_fd(tsk->evthr_t_fd)) {
		close(tsk->evthr_t_fd);
		tsk->evthr_t_fd = dns_tsk->dnstsk_fd = INVALID_SOCKET;
	}
/* needs locking if it should be done
//	dns_mgr_ctx->dnsmgr_tskstatus[dns_tsk->dnstsk_idx] = DNSTSK_ST_EOF;
*/
	return SM_SUCCESS;
}

/*
**  DNS_COMM_TSK -- DNS communication task: calls functions to perform
**		reads and writes.
**	This runs as task in the event thread system.
**
**	Parameters:
**		tsk -- evthr task
**
**	Return value:
**		usual sm_error code
*/

sm_ret_T
dns_comm_tsk(sm_evthr_task_P tsk)
{
	sm_ret_T ret;
	dns_tsk_P dns_tsk;

	SM_IS_EVTHR_TSK(tsk);
	dns_tsk = (dns_tsk_P) tsk->evthr_t_actx;
	SM_IS_DNSTSK(dns_tsk);
	MTA_LIBDNS_DBG_DPRINTF((smioerr,
		"sev=DBG, func=dns_comm, dns_tsk=%p, want=%x, got=%x\n",
		dns_tsk, tsk->evthr_t_rqevf, tsk->evthr_t_evocc));

	if (NULL == dns_tsk)
		return EVTHR_DEL;
	if (is_valid_fd(tsk->evthr_t_fd)) {
		ret = EVTHR_WAITQ;
		if (evthr_got_wr(tsk))
			ret = dns_tsk_wr(tsk);	/* XXX check ret here? */
		/* XXX This may turn off WR; change the order? */

		if (evthr_got_rd(tsk))
			ret = dns_tsk_rd(tsk);
		if (sm_error_perm(SM_EM_DNS, SM_E_EOF) == ret) {
			(void) dns_disconnect(tsk);

			/* this should only be done when needed */
			ret = dns_reconnect(tsk);
			return EVTHR_WAITQ;
			/* return EVTHR_WAITQ|evthr_r_no(EVTHR_EV_WR)|evthr_r_no(EVTHR_EV_RD); */
		}
		if (sm_is_err(ret)) {
			/* complain? do some cleanup? */
			return ret;
		}
		return ret;
	}
	return EVTHR_DEL;
}
